/* This file is part of Jabinit, copyright (C) 2005 Simon Rochester and John C. Morrison
*
* Jabinit is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* Jabinit is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.louisville.physics.jcm.jabinit;

import javax.swing.JApplet;
import javax.swing.UIManager;
import javax.swing.JTabbedPane;
import javax.swing.BoxLayout;
import javax.swing.JScrollPane;
import javax.swing.event.TableModelListener;
import javax.swing.event.TableModelEvent;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.IOException;
import edu.louisville.physics.jcm.utils.DeepCopy;

public class Jabinit extends JApplet implements ActionListener, TableModelListener {
	String connectStatus = null;
	JabinitIO jabinitIO;
	BandPlot bandPlot;
	JTabbedPane tabbedPane;

	CrystalData crystalData, oldCrystalData;
	CrystalPanel crystalPanel;
	CrystalControl crystalControl;

	//init sets up GUI
	public void init() {
		//set look and feel
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
	//		UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");
		} catch (Exception e) { }


		//get background color specified by caller, if exists
		String backgroundColorString = getParameter("backgroundColor");
		try {
			if (backgroundColorString != null) {
				getContentPane().setBackground(Color.decode(backgroundColorString));
			}
		} catch (NumberFormatException e) {	}

		//set layout manager
		getContentPane().setLayout(new BoxLayout(getContentPane(), BoxLayout.LINE_AXIS));
		//getRootPane().setBorder(BorderFactory.createRaisedBevelBorder());
	
		//create tabbed pane
		tabbedPane = new JTabbedPane();
		getContentPane().add(tabbedPane);
		tabbedPane.setAlignmentY(Component.BOTTOM_ALIGNMENT);
		
		//create crystal viewer
		crystalPanel = new CrystalPanel();
		tabbedPane.addTab("Crystal Structure", crystalPanel);
				
		//create band structure plot
		bandPlot = new BandPlot();
		tabbedPane.addTab("Band Structure", bandPlot);

		//control
		crystalControl = new CrystalControl(tabbedPane);
		getContentPane().add(crystalControl);
		crystalControl.setAlignmentY(Component.BOTTOM_ALIGNMENT);

		//create IO handler
		jabinitIO = new JabinitIO();
		tabbedPane.addTab("I/O",new JScrollPane(jabinitIO));	
	
		//add listeners
		crystalControl.runButton.addActionListener(this);
		crystalControl.resetButton.addActionListener(this);
		crystalControl.crystalSelector.addActionListener(this);
		crystalControl.primitiveVectorTable.getModel().addTableModelListener(this);
		crystalControl.atomsTable.getModel().addTableModelListener(this);
		crystalControl.plotPointsTable.getModel().addTableModelListener(this);
		crystalControl.miscParametersTable.getModel().addTableModelListener(this);
		
		//initialize crystal data
		validate();
		repaint();
		crystalControl.crystalSelector.setSelectedItem("Si");
	}
	
	//event handler
	public void actionPerformed(ActionEvent event) {
		Component source = (Component)event.getSource();
		try	{
			if (source == crystalControl.crystalSelector ||
				source == crystalControl.resetButton 
				) {
				crystalData = new CrystalData(
						(String)crystalControl.crystalSelector.getSelectedItem()
						);
		 		crystalPanel.displayCrystal(crystalData);
				crystalControl.setCrystalData(crystalData);
				updateRunButton();
			}
				
			if (source == crystalControl.runButton) {
				if (jabinitIO.getConnectionStatus() == null) {
					jabinitIO.connect(getCodeBase().getHost());
					showStatus(jabinitIO.getConnectionStatus());
				}
				jabinitIO.requestData(crystalData);
				bandPlot.plotData(jabinitIO.bandData);
				if (tabbedPane.getSelectedComponent() == crystalPanel)
					tabbedPane.setSelectedComponent(bandPlot);
				oldCrystalData = (CrystalData)DeepCopy.copy(crystalData);
				updateRunButton();
			}
		} catch (IOException e) {
			showStatus(e.toString());
		} catch (NullPointerException e) {
			showStatus(e.toString());
		}
	}
	
	public void tableChanged(TableModelEvent e) {
		crystalPanel.displayCrystal(crystalData);
		updateRunButton();
	}
	
	void updateRunButton() {
		crystalControl.runButton.setIcon(
			crystalData.equals(oldCrystalData) ? 
			crystalControl.blackRun : crystalControl.redRun
			);
	}
}
