/* This file is part of Jabinit, copyright (C) 2005 Simon Rochester and John C. Morrison
*
* Jabinit is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* Jabinit is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.louisville.physics.jcm.jabinit;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.StreamTokenizer;
import java.net.Socket;
import javax.swing.JTextArea;
import java.lang.reflect.Array;

public class JabinitIO extends JTextArea {
	Socket socket = null;
	PrintWriter writer = null;
	BufferedReader reader = null;
	StreamTokenizer tokenizer;
	String connectionStatus = null;

	public double[][] bandData;

	final static int PORT = 9097;

	public JabinitIO() {
		super(20,20);
		setEditable(false);
	}
	
	void connect(String host) throws IOException {
		socket = new Socket(host, PORT);
		writer = new PrintWriter(socket.getOutputStream(), true);
        reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		tokenizer = new StreamTokenizer(reader);
		connectionStatus = "Connected to " + socket.getInetAddress().getHostName() + ":" 
							+ socket.getPort();
	}

	public String getConnectionStatus() {
		return connectionStatus;
	}

	//cleanup
	void disconnect() throws IOException {
		writer.println("EXIT");
		if (socket != null) socket.close();
		if (writer != null) writer.close();
		if (reader != null) reader.close();
	}

	public void requestData(CrystalData crystalData) throws IOException, NullPointerException {
		
		setText("");
		
		//send number of atom types = # of pseudopotentials
		writer.println("" + crystalData.nAtomTypes);

		//send pseudopotential file names
		writer.println(crystalData.psp);

		//send input file text
		String inputText = 
			  "# Computation of the band structure.                                          \n"
            + "# First, a SCF density computation, then a non-SCF band structure calculation.\n"
            + "                                                                              \n"
            + "ndtset 2                                                                      \n"
            + "                                                                              \n"
            + "#Dataset 1 : usual self-consistent calculation                                \n"
            + "kptopt1 1             # Option for the automatic generation of k points,      \n"
            + "                      # taking into account the symmetry                      \n"
            + "nshiftk1 4                                                                    \n"
            + "shiftk1  0.5 0.5 0.5  # These shifts will be the same for all grids           \n"
            + "         0.5 0.0 0.0                                                          \n"
            + "         0.0 0.5 0.0                                                          \n"
            + "         0.0 0.0 0.5                                                          \n"
            + "ngkpt1  4 4 4                                                                 \n"
            + "prtden1  1            # Print the density, for use by dataset 2               \n"
            + "toldfe1  1.0d-6                                                               \n"
            + "                                                                              \n"
            + "#Dataset 2 : the band structure                                               \n"
            + "iscf2    -2                                                                   \n"
            + "getden2  -1                                                                   \n"
            + "kptopt2  -3                                                                   \n"
            + "nband2   8                                                                    \n"
            + "ndivk2   10 12 17    # 10, 12 and 17 divisions of the 3 segments, delimited   \n"
            + "                     # by 4 points.                                           \n"
            + "kptbounds2 \n"
			+ arrayToString(crystalData.bandPlotPoints, "\n")
            + "tolwfr2  1.0d-12                                                              \n"
            + "enunit2  1                # Will output the eigenenergies in eV               \n"
            + "                                                                              \n"
            + "                                                                              \n"
            + "#Definition of the unit cell                                                  \n"
            + "acell " + arrayToString(crystalData.acell, " ") + "\n"
            + "rprim                  # FCC primitive vectors (to be scaled by acell)        \n"
			+ arrayToString(crystalData.primitiveVectors, "\n")
            + "                                                                              \n"
            + "#Definition of the atom types                                                 \n"
            + "ntypat " + crystalData.nAtomTypes + "\n"
            + "znucl " + arrayToString(crystalData.zs, " ") + "\n"
            + "                                                                              \n"
            + "#Definition of the atoms                                                      \n"
            + "natom " + crystalData.nAtoms + "\n"
            + "typat " + arrayToString(crystalData.atomTypes, " ") + "\n"
            + "xred              # This keyword indicates that the location of the atoms     \n"
            + "                  # will follow, one triplet of numbers for each atom         \n"
            + arrayToString(crystalData.xred, "\n")
            + "                                                                              \n"
            + "#Definition of the planewave basis set                                        \n"
            + "ecut " + crystalData.ecut + " # Maximal kinetic energy cut-off, in Hartree    \n"
            + "                                                                              \n"
            + "#Definition of the SCF procedure                                              \n"
            + "nstep " + crystalData.maxCycles + " # Maximal number of SCF cycles            \n"
            + "diemac " + crystalData.diemac + "\n";
		
		append(inputText + "\n");
		writer.println(inputText);																						
		writer.println("END");																	
																								
		//read response until data set 2 is reached												
		String line;
		while (!(line = reader.readLine()).regionMatches(0, "== DATASET  2", 0, 13)) {
			append(line + "\n");
		}
		append(line + "\n");

		//read response until Eigenvalues of data set 2 are reached
		while (!(line = reader.readLine()).regionMatches(0, " Eigenvalues (   eV  ) for nkpt", 0, 31)) {
			append(line + "\n");
		}
		append(line + "\n");

		//get nkpt
		int nkpt = (new Integer(((line.split("nkpt=")[1]).split("k points")[0]).replaceAll(" ",""))).intValue();
		int nband = 8;
		
		//read band structure
		bandData = new double[nband][nkpt];
		for (int i=0; i<nkpt; i++) {
			append(reader.readLine() + "\n");
			for (int j=0; j<nband; j++) {
				tokenizer.nextToken();
				bandData[j][i] = tokenizer.nval;
			}
		}		

		//read response to the end
		while (!(line = reader.readLine()).equals("END")) {
			append(line + "\n");
		}
	}
	
	String arrayToString(Object a, String delimiter) {
		String s = "";
		for (int i=0; i<Array.getLength(a); i++)
			s += Array.get(a, i) + delimiter;
		return s;
	}
}
