/* $RCSfile: SmarterJmolAdapter.java,v $
 * $Author: migueljmol $
 * $Date: 2004/11/15 00:51:08 $
 * $Revision: 1.10 $
 *
 * Renamed and modified by Simon Rochester 2005/1/13
 * for Jabinit program
 *
 * Copyright (C) 2003-2004  The Jmol Development Team
 *
 * Contact: jmol-developers@lists.sf.net
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

package edu.louisville.physics.jcm.jmol;

import org.jmol.api.JmolAdapter;
import java.io.BufferedReader;
import edu.louisville.physics.jcm.jabinit.CrystalData;
import edu.louisville.physics.jcm.jabinit.Vector3D;


public class JabinitJmolAdapter extends JmolAdapter {
	CrystalData crystalData;	
  
  //////////////////////////////////////////////////////////////////
  // file related
  //////////////////////////////////////////////////////////////////

  public JabinitJmolAdapter(Logger logger) {
    super("JabinitJmolAdapter", logger);
  }

	public void setCrystalData(CrystalData crystalData) {
		this.crystalData = crystalData;
	}

  /**
   * Associate a clientFile object with a bufferedReader.
   * 
   * <p>Given the BufferedReader, return an object which represents the file
   * contents. The parameter <code>name</code> is assumed to be the
   * file name or URL which is the source of reader. Note that this 'file'
   * may have been automatically decompressed. Also note that the name
   * may be 'String', representing a string constant. Therefore, few
   * assumptions should be made about the <code>name</code> parameter.
   *
   * The return value is an object which represents a <code>clientFile</code>.
   * This <code>clientFile</code> will be passed back in to other methods.
   * If the return value is <code>instanceof String</code> then it is
   * considered an error condition and the returned String is the error
   * message. 
   *
   * @param name File name, String or URL acting as the source of the reader
   * @param bufferedReader The BufferedReader
   * @return The clientFile or String with an error message
   */
  public Object openBufferedReader(String name,
                                   BufferedReader bufferedReader) {
		return crystalData;
    
  }

  public String getFileTypeName(Object clientFile) { 
		return "Jabinit"; 
	}

  /**
   * Get the name of the atom set collection, if known.
   * 
   * <p>Some file formats contain a formal name of the molecule in the file.
   * If this method returns <code>null</code> then the JmolViewer will
   * automatically supply the file/URL name as a default.
   * @param clientFile
   * @return The atom set collection name or <code>null</code>
   */
  public String getAtomSetCollectionName(Object clientFile) { 
		return ((CrystalData)clientFile).name; 
	}

  /**
   * Get the estimated number of atoms contained in the file.
   *
   * <p>Just return -1 if you don't know (or don't want to figure it out)
   * @param clientFile The client file
   * @return The estimated number of atoms in the file
   */
  public int getEstimatedAtomCount(Object clientFile) {
  	return -1;
	}
  
  /**
   * Get the boolean whether coordinates are fractional.
   * @param clientFile The client file
   * @return true if the coordinates are fractional, default <code>false</code>
   */
  public boolean coordinatesAreFractional(Object clientFile) { 
		return true; 
	}

  /**
   * Get the notional unit cell.
   * 
   * <p>This method returns the parameters that define a crystal unitcell
   * the parameters are returned in a float[] in the following order
   * <code>a, b, c, alpha, beta, gamma</code>
   * <br><code>a, b, c</code> : angstroms
   * <br><code>alpha, beta, gamma</code> : degrees
   * <br>if there is no unit cell data then return null
   * @param clientFile The client file
   * @return The array of the values or <code>null</code>
   */
  public float[] getNotionalUnitcell(Object clientFile) { 
		CrystalData cd = (CrystalData)clientFile;
		return new float[] {
			cd.primitiveVectors[0].norm() * cd.acellAngstrom(0),
			cd.primitiveVectors[1].norm() * cd.acellAngstrom(1),
			cd.primitiveVectors[2].norm() * cd.acellAngstrom(2),
			cd.primitiveVectors[0].angle(cd.primitiveVectors[1]),
			cd.primitiveVectors[0].angle(cd.primitiveVectors[2]),
			cd.primitiveVectors[1].angle(cd.primitiveVectors[2]),
			};						
	}
  
  /**
   * Get an AtomIterator for retrieval of all atoms in the file.
   * 
   * <p>This method may not return <code>null</code>.
   * @param clientFile The client file
   * @return An AtomIterator
   * @see AtomIterator
   */
  public JmolAdapter.AtomIterator getAtomIterator(Object clientFile) {
    return new AtomIterator((CrystalData)clientFile);
  }
	
  /**
   * Get a BondIterator for retrieval of all bonds in the file.
   * 
   * <p>If this method returns <code>null</code> and no
   * bonds are defined then the JmolViewer will automatically apply its
   * rebonding code to build bonds between atoms.
   * @param clientFile The client file
   * @return A BondIterator or <code>null</code>
   * @see BondIterator
   */
  public BondIterator getBondIterator(Object clientFile) { 
		return null; 
	}

  /****************************************************************
   * AtomIterator is used to enumerate all the <code>clientAtom</code>
   * objects in a specified frame. 
   * Note that Java 1.1 does not have java.util.Iterator
   * so we will define our own AtomIterator
   ****************************************************************/
  class AtomIterator extends JmolAdapter.AtomIterator {
    CrystalData crystalData;
		int x,y,z,i,size;
		Vector3D bbMin,bbMax;
		
		AtomIterator(CrystalData crystalData) {
      this.crystalData = crystalData;
			size = 2*(int)crystalData.crystalSize;
			x = y = z = -size;
			bbMin = new Vector3D(0,0,0);
			bbMax = new Vector3D(crystalData.crystalSize);
    }
		
    public boolean hasNext() {
			do {
				if (++i == crystalData.nAtoms) {
					i = 0;			
					if (x++ == size) {
						x = -size;
						if (y++ == size) {
							y = -size;
							if (z++ == size) {
								return false;
							}
						}
					}
				}
			} while (!crystalData.getCartCoordinates(x,y,z,i).isInBox(bbMin,bbMax));
      return true;
    }

    public Object getUniqueID() { 
			return "" + x + y + z + i; 
		}
    
		public int getElementNumber() { 
			return crystalData.getAtomZ(i); 
		}
		
    public String getElementSymbol() { 
			return crystalData.getAtomName(i); 
		}
		
    public float getX() { 
			return x+crystalData.xred[i].x; 
		}

    public float getY() { 
			return y+crystalData.xred[i].y; 
		}
		
    public float getZ() { 
			return z+crystalData.xred[i].z; 
		}
  }
}
